//============================================================================//
//              PIC24 port bits access macros                                 //
// Original written by Ascold Volkov. Rewritten by Andy Mozzevilov            //
//  Corrected by Baser (21 Feb 2008)  Rewritten by Dmitiy Ogarenko (2009)     //
// Examples:                                                                  //
//  #define Pin1 D,2,H  ->     (Port, Bit number, Active level)               //
//  #define Pin2 C,3,L  ->     (Port, Bit number, Active level)               //
//  #define Pin2 C,3,OD ->     (Port, Bit number, openDrain)                  //
//  #define Pin2 C,3,O  ->     (Port, Bit number, Pull Up rezisror            //
//                              Toggle (change) output input/output)          //
// Summary:                                                                   //
//  ON(pin_x)                 - Set Active level on output pin_x              //
//  OFF(pin_x)                - Set Passive level on output pin_x             //
//  HIGH(pin_x)               - Set High level on output pin_x                //
//  LOW(pin_x)                - Set Low level on output pin_x                 //
//  OPEND_DRAIN_ON(x)         - Set open-drain output                         //
//  OPEND_DRAIN_OFF(x)        - Set digital output                            //
//  ANALOG_INPUT_ON(x)        - Pin configured in Analog mode                 //
//  ANALOG_INPUT_OFF(x)       - Pin  configured in Digital mode               //
//  SIGNAL(pin_x)             - Check Active level on input pin_x             //
//  LATCH(pin_x)              - Check Active level on output latch of pin_x   //
//  DIR_IS_IN(pin_x)          - Check if direction of pin_x is Input          //
//  DIR_IS_OUT(pin_x)         - Check if direction of pin_x is Output         //
//  TOGGLE(pin_x)             - Toggle (change) output pin_x level            //
//  BITNUM(pin_x)             - Return port pin_x number (0...15)             //
//  SET_DIR(pin_x, direction) - Set direction of pin_x (In or Out)            //
//============================================================================//
#ifndef _C30PORTMACROS_H
#define _C30PORTMACROS_H


#define _setL(port,bit)             BFA(BFA_WR, port, port##bit, 0)
#define _setH(port,bit)             BFA(BFA_WR, port, port##bit, 1)
#define _set(port,bit,val)          _set##val(LAT##port,bit)
#define _high(port,bit,val)         _setH(LAT##port,bit)
#define _openDrainOn(port,bit,val)  BFA(BFA_WR, ODC##port, OD##port##bit, 1)             
#define ON(x)                       _set(x)
#define HIGH(x)                     _high(x)
#define OPEND_DRAIN_ON(x)           _openDrainOn(x)
#define _analogInputOn(x)           BFA(BFA_WR, AD1PCFG,PCFG##x, 0)
#define ANALOG_INPUT_ON(val)        _analogInputOn(val);

#define _clrL(port,bit)             BFA(BFA_WR, port, port##bit, 1)
#define _clrH(port,bit)             BFA(BFA_WR, port, port##bit, 0)
#define _clr(port,bit,val)          _clr##val(LAT##port,bit)
#define _low(port,bit,val)          _clrH(LAT##port,bit)
#define _openDrainOff(port,bit,val) BFA(BFA_WR,ODC##port, OD##port##bit,  0)
#define OFF(x)                      _clr(x)
#define LOW(x)                      _low(x)
#define OPEND_DRAIN_OFF(x)          _openDrainOff(x)
#define ANALOG_INPUT_OFF(val)       BFA(BFA_WR, AD1PCFG,PCFG##val, 1);
//#define ANALOG_INPUT_OFF(val)       _analogInputOn;
//BFA(IPC0, INT0IP, BFA_RD);

#define _bitL(port,bit)             (!(BFA(BFA_RD, PORT##port, R##port##bit)))
#define _bitH(port,bit)             BFA(BFA_RD, PORT##port, R##port##bit)
#define _bit(port,bit,val)          _bit##val(port,bit)
#define _latchL(port,bit)           (!(BFA(BFA_RD, port, port##bit)))
#define _latchH(port,bit)            BFA(BFA_RD, port, port##bit)
#define _latch(port,bit,val)        _latch##val(LAT##port,bit)
#define _dirI(port,bit,val)         _latchH(TRIS##port,bit)
#define _dirO(port,bit,val)         _latchL(TRIS##port,bit)
#define SIGNAL(x)                   _bit(x)
#define LATCH(x)                    _latch(x)
#define DIR_IS_IN(x)                _dirI(x)
#define DIR_IS_OUT(x)               _dirO(x)
                                                                   
#define _xor(port,bit,val)          BFA(BFA_INV, LAT##port, LAT##port##bit)
#define TOGGLE(x)                   _xor(x)

#define _bitnum(port,bit,val)bit
#define BITNUM(x)                   _bitnum(x)

#define _setOut(port,bit)           BFA(BFA_WR, port, port##bit, 0)
#define _setIn(port,bit)            BFA(BFA_WR, port, port##bit, 1)
#define _mode(port,bit,val,mode)    _set##mode(TRIS##port,bit)
//mode = Out or In
#define SET_DIR(x,mode)             _mode(x,mode)

//tst_brgh(val, MODE)									BFA(BFA_RD, U##val##MODE, BRGH)

#define TST_BRGH(uartN)             BFA(BFA_RD, U##uartN##MODE, BRGH)
#define _setBrg(valBrg , uartN)			if (TST_BRGH(uartN)) 						\
																			U2BRG = (FREQ) / 2 /16/valBrg -1; 										\
																				else 												\
																					U2BRG = (FREQ) / 2 /16/valBrg -1;
#define SET_BRGH(x)             		_setBrg(x)

#define DISI(n) \
asm volatile ("disi #%0" \
: /* no outputs */ \
: "i" (n))



#endif /* _C30PORTMACROS_H */
